import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;

import java.awt.geom.GeneralPath;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;

import java.util.Enumeration;

import javax.media.j3d.*;

import com.sun.j3d.exp.swing.impl.AutoOffScreenCanvas3D;
import com.sun.j3d.utils.behaviors.mouse.MouseRotate;
import com.sun.j3d.utils.behaviors.mouse.MouseTranslate;
import com.sun.j3d.utils.behaviors.mouse.MouseWheelZoom;
import com.sun.j3d.utils.behaviors.mouse.MouseZoom;

import javax.vecmath.Point3f;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector4f;

import javax.swing.JFrame;
import javax.swing.JPanel;

/**
 * 2007/01/14
 *
 * @version 1.0
 * @since Java 3D 1.5.0
 * @author August Lammersdorf, www.InteractiveMesh.com
 */
public class SceneInScene {
    
    private BoundingSphere          globalBounds    =   null;
 
    private View                    viewOffScreen   =   null;
    
    private Background              bgTarget        =   null;
    
    private Raster                  rasterTarget    =   null;
    private TransparencyAttributes  rasterTransAttr =   null;
    
    private Texture2D               texture2DTarget =   null;
    private TextureAttributes       texAttrTarget   =   null;
    
    private int                     monitorWidth    =   256;
    private int                     monitorHeight   =   256;

    private SceneCapture            sceneCapture    =   null;
    private MonitorUpdater          monitorUpdater  =   null;
    
    
    public static void main(final String args[]) {
        
    	// Set Java 3D properties at first
        // System.setProperty("j3d.rend", "ogl"); // "d3d" "jogl" "ogl"
        // System.setProperty("j3d.transparentOffScreen", "true"); // true
        // System.setProperty("j3d.usePbuffer", "true"); // false
        
        System.out.println("SceneInScene-1.0 is running.");

        new SceneInScene();
    }

    public SceneInScene() {
        
        // Bounds
        globalBounds = new BoundingSphere();
        globalBounds.setRadius(Double.MAX_VALUE);		

        // SceneCapture: an auto-off-screen Canvas3D
        GraphicsConfigTemplate3D gct = new GraphicsConfigTemplate3D();
        gct.setDoubleBuffer(GraphicsConfigTemplate3D.UNNECESSARY);
        gct.setStereo(GraphicsConfigTemplate3D.UNNECESSARY);
        GraphicsConfiguration gc = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getBestConfiguration(gct);;
        sceneCapture = new SceneCapture(gc);
        // Set size and transparency (set j3d.transparentOffScreen adequately)
        sceneCapture.setUp(new Dimension(monitorWidth, monitorHeight), false);
                
        // Source world   	
        createSourceWorld(); 
        // Target world
        createTargetWorld(); 

        // Add auto-off-screen Canvas3D as scene capture to source world
        viewOffScreen.addCanvas3D(sceneCapture);

        // Place a shared capture's monitor into target world
        ImageComponent2D monitor = sceneCapture.getMonitor();
        
        // Used as Background image
        bgTarget.setImage(monitor);
        bgTarget.setImageScaleMode(Background.SCALE_NONE); // SCALE_REPEAT
        
        // Used as Texture2D image
        texture2DTarget.setImage(0, monitor);
        texAttrTarget.setTextureMode(TextureAttributes.MODULATE);
        
        // Used as Raster image
        rasterTarget.setImage(monitor);  
        rasterTransAttr.setTransparencyMode(TransparencyAttributes.NICEST);
        rasterTransAttr.setTransparency(0.0f);
                       
        // Start broadcasting
        monitorUpdater.setEnable(true);
    }
    
    // Monitors in the target world will be refreshed each time a frame is rendered in the source world 
    private class MonitorUpdater extends Behavior {

        private WakeupOnElapsedFrames passiveFrames = new WakeupOnElapsedFrames(0, true);		
        
        @Override
        public void initialize() {
            wakeupOn(passiveFrames);
        }
        @Override
        public void processStimulus(Enumeration criteria) {	
        	
            sceneCapture.updateMonitor();           
            
            wakeupOn(passiveFrames);
        }
    }
    /*
     NOTE: AutoOffScreenCanvas3D is an experimental interface, which is not intended for use by applications.
    */
    public class SceneCapture extends Canvas3D implements AutoOffScreenCanvas3D, 
                                                          ImageComponent2D.Updater {
        
	// physical screen width : 0.0254/90.0 * screen width (in pixels)
	// physical screen height : 0.0254/90.0 * screen height (in pixels)
	private	float               pixSize         =   0.0254f/90.0f * 1.1f;
        
        private int                 width           =   1;
        private int                 height          =   1;

        private ImageComponent2D    sceneMonitor    =   null;
        
        private int[]               c3dImageData    =   null;
        private int[]               i2dImageData    =   null;
        
        public SceneCapture(GraphicsConfiguration gc) {
            super(gc, true);
        }
        
        public ImageComponent2D getMonitor() {
            return sceneMonitor;
        }
    
        public void setUp(Dimension size, boolean alpha) {

            if (size.width < 1 || size.height < 1)
                throw new IllegalArgumentException("SceneCapture: width or height are not positive");

            width = size.width;
            height = size.height;
            
            Screen3D offScreen3D = this.getScreen3D();
            offScreen3D.setSize(size);				 
            offScreen3D.setPhysicalScreenWidth(width * pixSize);
            offScreen3D.setPhysicalScreenHeight(height * pixSize);	

            int imageFormat = ImageComponent2D.FORMAT_RGB;
            int imageType = BufferedImage.TYPE_INT_RGB;

            if (alpha) {
                imageFormat = ImageComponent2D.FORMAT_RGBA;
                imageType = BufferedImage.TYPE_INT_ARGB;
            }

            try {
                // Off-screnn buffer
                BufferedImage c3dImage = new BufferedImage(width, height, imageType);
                c3dImageData = ((DataBufferInt)c3dImage.getRaster().getDataBuffer()).getData();
                // byReference, !yUp
                ImageComponent2D ic2D = new ImageComponent2D(imageFormat, c3dImage, true, false);

                setOffScreenBuffer(ic2D);	
                
                // Monitor image
                BufferedImage i2dImage = new BufferedImage(width, height, imageType);
                i2dImageData = ((DataBufferInt)i2dImage.getRaster().getDataBuffer()).getData();
                // byReference, !yUp
                sceneMonitor = new ImageComponent2D(imageFormat, i2dImage, true, false);
                sceneMonitor.setCapability(ImageComponent.ALLOW_IMAGE_WRITE);
                
                // Alternative: Share a single BufferedImage for both, off-screen buffer and monitor,
                //              to avoid copying image data in 'this.updateData()'
            }
            catch (IllegalStateException e) {
                System.exit(0);
            }
            catch (RestrictedAccessException e) {
                System.exit(0);
            }
            catch (IllegalSharingException e) {
                System.exit(0);
            }
            catch (IllegalArgumentException e) {
                System.exit(0);
            }
        }
	        
        // Called from MonitorUpdater
        public void updateMonitor() {
            // ByReference update
            sceneMonitor.updateData(this, 0, 0, width, height);
        }
        
        // Interface ImageComponent2D.Updater
        public void updateData(ImageComponent2D imageComponent, int x, int y, int width, int height) {
            System.arraycopy( c3dImageData, 0, i2dImageData, 0, width*height);
        }
    }
    
    private void createSourceWorld() {		

        //
        // Viewing
        //
        PhysicalBody phB = new PhysicalBody();
        PhysicalEnvironment phE = new PhysicalEnvironment();

        /* TODO: comment out this on-screen View */
        View view = new View();
        view.setPhysicalBody(phB);
        view.setPhysicalEnvironment(phE);
        
        // Off-screen View
        viewOffScreen = new View();
        viewOffScreen.setPhysicalBody(phB);
        viewOffScreen.setPhysicalEnvironment(phE);
        
        viewOffScreen.setFrontClipDistance(0.01);
        
        /* TODO: comment out this on-screen Canvas3D */
        GraphicsConfigTemplate3D gCT = new GraphicsConfigTemplate3D();
        GraphicsConfiguration gcfg = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getBestConfiguration(gCT);
        Canvas3D canvas3D = null;
        try {
            canvas3D = new Canvas3D(gcfg);
        }
        catch (NullPointerException e) {
            System.exit(0);
        }
        catch (IllegalArgumentException e) {
            System.exit(0);
        }        
        //
        // SuperStructure
        //
        VirtualUniverse vu = new VirtualUniverse();
        Locale locale = new Locale(vu);
        
        //
        // BranchGraphs
        //		
        BranchGroup sceneBranch = new BranchGroup();
        BranchGroup viewBranch = new BranchGroup();
        BranchGroup enviBranch = new BranchGroup();
        
        // ViewBranch
        TransformGroup viewTG = new TransformGroup();
        Transform3D viewTransform = new Transform3D();
        viewTransform.setTranslation(new Vector3f(0.0f, 0.35f, 2.0f));
        viewTG.setTransform(viewTransform);

        ViewPlatform vp = new ViewPlatform();

        DirectionalLight headLight = new DirectionalLight();
        headLight.setInfluencingBounds(globalBounds);

        viewTG.addChild(vp);
        viewTG.addChild(headLight);
        viewBranch.addChild(viewTG);
        
        /* TODO: comment out this on-screen View */
        view.attachViewPlatform(vp);
        // 
        viewOffScreen.attachViewPlatform(vp);
        
        // EnviBranch
        Background bg = new Background();
        bg.setCapability(Background.ALLOW_IMAGE_WRITE);
        bg.setApplicationBounds(globalBounds);
        bg.setColor(0.0f, 1.0f, 1.0f);
        enviBranch.addChild(bg);
        
        monitorUpdater = new MonitorUpdater();
        monitorUpdater.setSchedulingBounds(globalBounds);
        monitorUpdater.setEnable(false);
        enviBranch.addChild(monitorUpdater);

        // SceneBranch
        // Target MouseNavigation
        TransformGroup mouseTG = new TransformGroup();
        mouseTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        // Target RotationInterpoaltor
        TransformGroup rotInterpTG = new TransformGroup();
        rotInterpTG.setCapability(Group.ALLOW_CHILDREN_EXTEND);
        rotInterpTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        // MouseNavigation on top of RotationInterpoaltor
        mouseTG.addChild(rotInterpTG);	               
        sceneBranch.addChild(mouseTG);

        // MouseRotate
        MouseRotate mouseRotate = new MouseRotate(canvas3D, mouseTG);
        mouseRotate.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(mouseRotate);
        // MouseTranslate 
        MouseTranslate mouseTrans = new MouseTranslate(canvas3D, mouseTG);
        mouseTrans.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(mouseTrans);
        // MouseWheelZoom 
        MouseWheelZoom mouseWZoom = new MouseWheelZoom(canvas3D, mouseTG);
        mouseWZoom.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(mouseWZoom);
        // MouseZoom
        MouseZoom mouseZoom = new MouseZoom(canvas3D, mouseTG);
        mouseZoom.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(mouseZoom);

        // RotationInterpolator: starts at once, stops never
        Alpha alpha = new Alpha();
        alpha.setIncreasingAlphaDuration(6000);        
        
        RotationInterpolator rotInterp = new RotationInterpolator(alpha, rotInterpTG); 
        rotInterp.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(rotInterp);
        
        // Shape "J"
        
        // Appearance
        Appearance appearance = new Appearance();

        Material material = new Material();
        material.setDiffuseColor(0.0f, 0.0f, 1.0f);
        material.setSpecularColor(0.0f, 0.0f, 0.5f);
        material.setShininess(128);

        appearance.setMaterial(material);

        // Geometry 		
        GeneralPath shape = new GeneralPath();
        shape.moveTo(0.0f, 0.0f);
        shape.lineTo(0.2f, 0.0f);	

        Font font = new Font("Dialog", Font.PLAIN, 1);
        Font3D font3D = new Font3D(font, 0.005, new FontExtrusion(shape));

        char[] chars = Character.toChars(74);

        TriangleArray triangleArrayJ = (TriangleArray)font3D.getGlyphGeometry(chars[0]);

        // Shape3D		
        Shape3D shape3D = new Shape3D(triangleArrayJ, appearance);		

        rotInterpTG.addChild(shape3D);
        
        // Set live
        locale.addBranchGraph(sceneBranch);
        locale.addBranchGraph(viewBranch);
        locale.addBranchGraph(enviBranch);

        /* TODO: comment out this on-screen JFrame */
        view.addCanvas3D(canvas3D);
    	
        JPanel jPanel = new JPanel(new BorderLayout());
        jPanel.setPreferredSize(new Dimension(512, 512));
        jPanel.add(canvas3D);

        JFrame jFrame = new JFrame();
        jFrame.setTitle("SceneInScene Source");
        jFrame.add(jPanel);
        jFrame.pack();
        jFrame.setVisible(true);       
    }
    
    private void createTargetWorld() {		
        //
        // Viewing
        //
        View view = new View();
        PhysicalBody phB = new PhysicalBody();
        PhysicalEnvironment phE = new PhysicalEnvironment();
        view.setPhysicalBody(phB);
        view.setPhysicalEnvironment(phE);
        
        GraphicsConfigTemplate3D gCT = new GraphicsConfigTemplate3D();
        GraphicsConfiguration gcfg = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getBestConfiguration(gCT);
        Canvas3D canvas3D = null;
        try {
            canvas3D = new Canvas3D(gcfg);
        }
        catch (NullPointerException e) {
            System.exit(0);
        }
        catch (IllegalArgumentException e) {
            System.exit(0);
        }        

        //
        // SuperStructure
        //
        VirtualUniverse vu = new VirtualUniverse();
        Locale locale = new Locale(vu);
        
        //
        // BranchGraphs
        //		
        BranchGroup sceneBranch = new BranchGroup();
        BranchGroup viewBranch = new BranchGroup();
        BranchGroup enviBranch = new BranchGroup();
        
        // ViewBranch
        TransformGroup viewTG = new TransformGroup();
        Transform3D viewTransform = new Transform3D();
        viewTransform.setTranslation(new Vector3f(0.0f, 0.0f, 3.5f));
        viewTG.setTransform(viewTransform);

        ViewPlatform vp = new ViewPlatform();

        DirectionalLight headLight = new DirectionalLight();
        headLight.setInfluencingBounds(globalBounds);

        viewTG.addChild(vp);
        viewTG.addChild(headLight);
        viewBranch.addChild(viewTG);
        
        view.attachViewPlatform(vp);
        
        // EnviBranch
        bgTarget = new Background();
        bgTarget.setCapability(Background.ALLOW_IMAGE_WRITE);
        bgTarget.setCapability(Background.ALLOW_IMAGE_SCALE_MODE_WRITE);
        bgTarget.setApplicationBounds(globalBounds);
        bgTarget.setColor(0.0f, 1.0f, 0.8f);
        enviBranch.addChild(bgTarget);

        // SceneBranch
        // Target MouseNavigation
        TransformGroup mouseTG = new TransformGroup();
        mouseTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        Transform3D mTransform = new Transform3D();
        mTransform.setEuler(new Vector3d(0.0, -0.6, 0.0));
        mouseTG.setTransform(mTransform);
        sceneBranch.addChild(mouseTG);
        
        // MouseRotate 
        MouseRotate mouseRotate = new MouseRotate(canvas3D, mouseTG);
        mouseRotate.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(mouseRotate);
        // MouseTranslate 
        MouseTranslate mouseTrans = new MouseTranslate(canvas3D, mouseTG);
        mouseTrans.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(mouseTrans);
        // MouseWheelZoom
        MouseWheelZoom mouseWZoom = new MouseWheelZoom(canvas3D, mouseTG);
        mouseWZoom.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(mouseWZoom);
        // MouseZoom 
        MouseZoom mouseZoom = new MouseZoom(canvas3D, mouseTG);
        mouseZoom.setSchedulingBounds(globalBounds);
        sceneBranch.addChild(mouseZoom);


        // Raster 
        Appearance rAppearance = new Appearance();
        
        rasterTransAttr = new TransparencyAttributes();
        rasterTransAttr.setCapability(TransparencyAttributes.ALLOW_MODE_WRITE);
        rasterTransAttr.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
        rAppearance.setTransparencyAttributes(rasterTransAttr);
        
        rasterTarget = new Raster();
        rasterTarget.setCapability(Raster.ALLOW_IMAGE_WRITE);
        rasterTarget.setClipMode(Raster.CLIP_IMAGE);
        rasterTarget.setPosition(new Point3f(1.0f, 0.0f, 0.0f));
        rasterTarget.setSize(monitorWidth, monitorHeight);   
        rasterTarget.setImage(new ImageComponent2D(ImageComponent2D.FORMAT_RGB, 
                new BufferedImage(monitorWidth, monitorHeight, BufferedImage.TYPE_INT_RGB)));
        
        Shape3D shapeRaster = new Shape3D();
        shapeRaster.setGeometry(rasterTarget);
        shapeRaster.setAppearance(rAppearance);
        mouseTG.addChild(shapeRaster);
        
        // Box
        // Appearance
        Appearance bAppearance = new Appearance();
        Material material = new Material();
        material.setDiffuseColor(1.0f, 1.0f, 1.0f);
        material.setSpecularColor(0.0f, 0.0f, 0.0f);
        material.setShininess(128);
        bAppearance.setMaterial(material);
        
        texture2DTarget = new Texture2D(Texture2D.BASE_LEVEL, Texture.RGBA, monitorWidth, monitorHeight);
	texture2DTarget.setCapability(Texture.ALLOW_IMAGE_WRITE);
        texture2DTarget.setImage(0, new ImageComponent2D(ImageComponent2D.FORMAT_RGBA, 
                new BufferedImage(monitorWidth, monitorHeight, BufferedImage.TYPE_INT_ARGB)));
                
	texAttrTarget = new TextureAttributes();
        texAttrTarget.setCapability(TextureAttributes.ALLOW_MODE_WRITE);
        
        TexCoordGeneration texCoordGen = new TexCoordGeneration();
        texCoordGen.setPlaneS(new Vector4f(1.0f, 0.0f, 0.0f, 0.5f));
        texCoordGen.setPlaneT(new Vector4f(0.0f, 1.0f, 0.0f, 0.5f));
        
        TextureUnitState[] texUnitStates = new TextureUnitState[1];
        texUnitStates[0] = new TextureUnitState(texture2DTarget, texAttrTarget, texCoordGen);

	bAppearance.setTextureUnitState(texUnitStates);

        // Geometry
    	float[] coords = {
            // Front
            -0.5f, -0.5f,  0.5f,
             0.5f, -0.5f,  0.5f,
             0.5f,  0.5f,  0.5f,
            -0.5f,  0.5f,  0.5f,
            // Back
             0.5f, -0.5f, -0.5f,
            -0.5f, -0.5f, -0.5f,           
            -0.5f,  0.5f, -0.5f,
             0.5f,  0.5f, -0.5f};   	
    	int[] coordIndices = {
            0, 1, 2, 3,   4, 5, 6, 7,  
            1, 4, 7, 2,   5, 0, 3, 6,
            3, 2, 7, 6,   0, 5, 4, 1};    	
    	float[] normals = {
            0.0f, 0.0f, 1.0f,    0.0f,  0.0f, -1.0f,   
            1.0f, 0.0f, 0.0f,   -1.0f,  0.0f,  0.0f,   
            0.0f, 1.0f, 0.0f,    0.0f, -1.0f,  0.0f};   			
    	int[] normalsIndices = {
            0, 0, 0, 0,   1, 1, 1, 1, 
            2, 2, 2, 2,   3, 3, 3, 3, 
            4, 4, 4, 4,   5, 5, 5, 5};
    	
        IndexedQuadArray geometry = new IndexedQuadArray(
                    8, GeometryArray.COORDINATES | GeometryArray.NORMALS, 24);
        geometry.setCoordinates(0, coords);
        geometry.setCoordinateIndices(0, coordIndices);
        geometry.setNormals(0, normals);
        geometry.setNormalIndices(0, normalsIndices);

        Shape3D boxShape = new Shape3D(geometry, bAppearance);
        
        mouseTG.addChild(boxShape);

        // Set live
        locale.addBranchGraph(sceneBranch);
        locale.addBranchGraph(viewBranch);
        locale.addBranchGraph(enviBranch);

        // Show frame
    	view.addCanvas3D(canvas3D);
        
        JPanel jPanel = new JPanel(new BorderLayout());
        jPanel.setPreferredSize(new Dimension(512, 512));
        jPanel.add(canvas3D);

        JFrame jFrame = new JFrame();
        jFrame.setTitle("SceneInScene Target");
        jFrame.setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        jFrame.setLocation(550, 0);
        jFrame.add(jPanel);
        jFrame.pack();
        jFrame.setVisible(true);
    }
}
